@@ -60,7 +60,7 @@ bool buffer_track_alloc = get_env_bool("CEPH_BUFFER_TRACK");
     unsigned len;
     atomic_t nref;
 
-    raw(unsigned l) : len(l), nref(0)
+    raw(unsigned l) : data(NULL), len(l), nref(0)
     { }
     raw(char *c, unsigned l) : data(c), len(l), nref(0)
     { }
@@ -88,10 +88,13 @@ bool buffer_track_alloc = get_env_bool("CEPH_BUFFER_TRACK");
   class buffer::raw_malloc : public buffer::raw {
   public:
     raw_malloc(unsigned l) : raw(l) {
-      if (len)
+      if (len) {
 	data = (char *)malloc(len);
-      else
+        if (!data)
+          throw bad_alloc();
+      } else {
 	data = 0;
+      }
       inc_total_alloc(len);
       bdout << "raw_malloc " << this << " alloc " << (void *)data << " " << l << " " << buffer::get_total_alloc() << bendl;
     }
@@ -378,6 +381,17 @@ bool buffer_track_alloc = get_env_bool("CEPH_BUFFER_TRACK");
     return 0;
   }
 
+  bool buffer::ptr::is_zero() const
+  {
+    const char *data = c_str();
+    for (size_t p = 0; p < _len; p++) {
+      if (data[p] != 0) {
+	return false;
+      }
+    }
+    return true;
+  }
+
   void buffer::ptr::append(char c)
   {
     assert(_raw);
@@ -634,15 +648,45 @@ bool buffer_track_alloc = get_env_bool("CEPH_BUFFER_TRACK");
   {
     if (length() != other.length())
       return false;
-    bufferlist::iterator me = begin();
-    bufferlist::iterator him = other.begin();
-    while (!me.end()) {
-      if (*me != *him)
-        return false;
-      ++me;
-      ++him;
+
+    // buffer-wise comparison
+    if (true) {
+      std::list<ptr>::const_iterator a = _buffers.begin();
+      std::list<ptr>::const_iterator b = other._buffers.begin();
+      unsigned aoff = 0, boff = 0;
+      while (a != _buffers.end()) {
+	unsigned len = a->length() - aoff;
+	if (len > b->length() - boff)
+	  len = b->length() - boff;
+	if (memcmp(a->c_str() + aoff, b->c_str() + boff, len) != 0)
+	  return false;
+	aoff += len;
+	if (aoff == a->length()) {
+	  aoff = 0;
+	  a++;
+	}
+	boff += len;
+	if (boff == b->length()) {
+	  boff = 0;
+	  b++;
+	}
+      }
+      assert(b == other._buffers.end());
+      return true;
+    }
+
+    // byte-wise comparison
+    if (false) {
+      bufferlist::iterator me = begin();
+      bufferlist::iterator him = other.begin();
+      while (!me.end()) {
+	if (*me != *him)
+	  return false;
+	++me;
+	++him;
+      }
+      return true;
     }
-    return true;
   }
 
   bool buffer::list::is_page_aligned() const
@@ -665,6 +709,17 @@ bool buffer_track_alloc = get_env_bool("CEPH_BUFFER_TRACK");
     return true;
   }
 
+  bool buffer::list::is_zero() const {
+    for (std::list<ptr>::const_iterator it = _buffers.begin();
+	 it != _buffers.end();
+	 it++) {
+      if (!it->is_zero()) {
+	return false;
+      }
+    }
+    return true;
+  }
+
   void buffer::list::zero()
   {
     for (std::list<ptr>::iterator it = _buffers.begin();
@@ -771,6 +826,14 @@ void buffer::list::rebuild_page_aligned()
     bl.last_p = bl.begin();
   }
 
+  void buffer::list::claim_prepend(list& bl)
+  {
+    // steal the other guy's buffers
+    _len += bl._len;
+    _buffers.splice( _buffers.begin(), bl._buffers );
+    bl._len = 0;
+    bl.last_p = bl.begin();
+  }
 
   void buffer::list::copy(unsigned off, unsigned len, char *dest) const
   {
@@ -1207,6 +1270,8 @@ int buffer::list::write_fd(int fd) const
 
 void buffer::list::hexdump(std::ostream &out) const
 {
+  std::ios_base::fmtflags original_flags = out.flags();
+
   out.setf(std::ios::right);
   out.fill('0');
 
@@ -1232,7 +1297,8 @@ void buffer::list::hexdump(std::ostream &out) const
     }
     out << std::dec << std::endl;
   }
-  out.unsetf(std::ios::right);
+
+  out.flags(original_flags);
 }
 
 std::ostream& operator<<(std::ostream& out, const buffer::raw &r) {
