@@ -118,6 +118,24 @@ public:
     crush->choose_total_tries = n;
   }
 
+  int get_chooseleaf_descend_once() {
+    return crush->chooseleaf_descend_once;
+  }
+  void set_chooseleaf_descend_once(int n) {
+    crush->chooseleaf_descend_once = !!n;
+  }
+
+  bool has_nondefault_tunables() const {
+    return
+      (crush->choose_local_tries != 2 ||
+       crush->choose_local_fallback_tries != 5 ||
+       crush->choose_total_tries != 19);
+  }
+  bool has_nondefault_tunables2() const {
+    return
+      crush->chooseleaf_descend_once != 0;
+  }
+
   // bucket types
   int get_num_type_names() const {
     return type_map.size();
@@ -145,19 +163,25 @@ public:
   // item/bucket names
   bool name_exists(const char *s) {
     string name(s);
+    return name_exists(name);
+  }
+  bool name_exists(const string& name) {
     build_rmaps();
     return name_rmap.count(name);
   }
   bool item_exists(int i) {
     return name_map.count(i);
   }
-  int get_item_id(const char *s) {
-    string name(s);
+  int get_item_id(const string& name) {
     build_rmaps();
     if (name_rmap.count(name))
       return name_rmap[name];
     return 0;  /* hrm */
   }
+  int get_item_id(const char *s) {
+    string name(s);
+    return get_item_id(name);
+  }
   const char *get_item_name(int t) const {
     std::map<int,string>::const_iterator p = name_map.find(t);
     if (p != name_map.end())
@@ -166,6 +190,9 @@ public:
   }
   void set_item_name(int i, const char *n) {
     string name(n);
+    set_item_name(i, name);
+  }
+  void set_item_name(int i, const string& name) {
     name_map[i] = name;
     if (have_rmaps)
       name_rmap[name] = i;
@@ -195,6 +222,15 @@ public:
 
   void find_roots(set<int>& roots) const;
 
+  /**
+   * see if an item is contained within a subtree
+   *
+   * @param root haystack
+   * @param item needle
+   * @return true if the item is located beneath the given node
+   */
+  bool subtree_contains(int root, int item) const;
+
   /**
    * see if item is located where we think it is
    *
@@ -212,8 +248,8 @@ public:
    * @param weight optional pointer to weight of item at that location
    * @return true if item is at specified location
    */
-  bool check_item_loc(CephContext *cct, int item, map<string,string>& loc, int *iweight);
-  bool check_item_loc(CephContext *cct, int item, map<string,string>& loc, float *weight) {
+  bool check_item_loc(CephContext *cct, int item, const map<string,string>& loc, int *iweight);
+  bool check_item_loc(CephContext *cct, int item, const map<string,string>& loc, float *weight) {
     int iweight;
     bool ret = check_item_loc(cct, item, loc, &iweight);
     if (weight)
@@ -275,7 +311,7 @@ public:
    * @param loc location (map of type to bucket names)
    * @return 0 for success, negative on error
    */
-  int insert_item(CephContext *cct, int id, float weight, string name, map<string,string>& loc);
+  int insert_item(CephContext *cct, int id, float weight, string name, const map<string,string>& loc);
 
   /**
    * move a bucket in the hierarchy to the given location
@@ -288,7 +324,7 @@ public:
    * @param loc location (map of type to bucket names)
    * @return 0 for success, negative on error
    */
-  int move_bucket(CephContext *cct, int id, map<string,string>& loc);
+  int move_bucket(CephContext *cct, int id, const map<string,string>& loc);
 
   /**
    * add or update an item's position in the map
@@ -303,7 +339,20 @@ public:
    * @param loc location (map of type to bucket names)
    * @return 0 for no change, 1 for successful change, negative on error
    */
-  int update_item(CephContext *cct, int id, float weight, string name, map<string,string>& loc);
+  int update_item(CephContext *cct, int id, float weight, string name, const map<string,string>& loc);
+
+  /**
+   * create or move an item, but do not adjust its weight if it already exists
+   *
+   * @param cct cct
+   * @param item item id
+   * @param weight initial item weight (if we need to create it)
+   * @param name item name
+   * @param loc location (map of type to bucket names)
+   * @return 0 for no change, 1 for successful change, negative on error
+   */
+  int create_or_move_item(CephContext *cct, int item, float weight, string name,
+			  const map<string,string>& loc);
 
   /**
    * remove an item from the map
@@ -313,6 +362,21 @@ public:
    * @return 0 on success, negative on error
    */
   int remove_item(CephContext *cct, int id);
+
+  /**
+   * get an item's weight
+   *
+   * Will return the weight for the first instance it finds.
+   *
+   * @param cct cct
+   * @param id item id to check
+   * @return weight of item
+   */
+  int get_item_weight(int id);
+  float get_item_weightf(int id) {
+    return (float)get_item_weight(id) / (float)0x10000;
+  }
+
   int adjust_item_weight(CephContext *cct, int id, int weight);
   int adjust_item_weightf(CephContext *cct, int id, float weight) {
     return adjust_item_weight(cct, id, (int)(weight * (float)0x10000));
@@ -492,8 +556,12 @@ private:
     // get the parent bucket
     crush_bucket *parent_bucket = get_bucket(parent_id);
 
-    // remove the bucket from the parent
-    crush_bucket_remove_item(parent_bucket, item);
+    if (!IS_ERR(parent_bucket)) {
+      // remove the bucket from the parent
+      crush_bucket_remove_item(parent_bucket, item);
+    } else if (PTR_ERR(parent_bucket) != -ENOENT) {
+      return PTR_ERR(parent_bucket);
+    }
 
     // check that we're happy
     int test_weight = 0;
