@@ -124,4 +124,19 @@ void crush_destroy(struct crush_map *map)
 	kfree(map);
 }
 
+// methods to check for safe arithmetic operations
+int crush_addition_is_unsafe(__u32 a, __u32 b)
+{
+  if ((((__u32)(-1)) - b) < a)
+    return 1;
+  else
+    return 0;
+}
 
+int crush_multiplication_is_unsafe(__u32  a, __u32 b)
+{
+  if ((((__u32)(-1)) / b) < a)
+    return 1;
+  else
+    return 0;
+}
