@@ -46,6 +46,7 @@
 #include "messages/MMonCommand.h"
 
 #include <memory>
+#include <errno.h>
 
 
 #define CEPH_MON_PROTOCOL     9 /* cluster internal */
@@ -249,7 +250,7 @@ public:
   void update_logger();
 
   // -- paxos -- These vector indices are matched
-  vector<Paxos*> paxos;
+  list<Paxos*> paxos;
   vector<PaxosService*> paxos_service;
 
   Paxos *add_paxos(int type);
@@ -368,11 +369,12 @@ public:
   
   void forward_request_leader(PaxosServiceMessage *req);
   void handle_forward(MForward *m);
-  void try_send_message(Message *m, entity_inst_t to);
+  void try_send_message(Message *m, const entity_inst_t& to);
   void send_reply(PaxosServiceMessage *req, Message *reply);
   void no_reply(PaxosServiceMessage *req);
   void resend_routed_requests();
   void remove_session(MonSession *s);
+  void remove_all_sessions();
 
   void send_command(const entity_inst_t& inst,
 		    const vector<string>& com, version_t version);
@@ -392,7 +394,9 @@ public:
     void finish(int r) {
       if (r >= 0)
 	mon->reply_command(m, rc, rs, rdata, version);
-      else
+      else if (r == -ECANCELED) {
+	m->put();
+      } else
 	mon->_ms_dispatch(m);
     }
   };
@@ -404,7 +408,10 @@ public:
   public:
     C_RetryMessage(Monitor *m, Message *ms) : mon(m), msg(ms) {}
     void finish(int r) {
-      mon->_ms_dispatch(msg);
+      if (r == -ECANCELED) {
+	msg->put();
+      } else
+	mon->_ms_dispatch(msg);
     }
   };
 
@@ -422,7 +429,7 @@ public:
   bool ms_get_authorizer(int dest_type, AuthAuthorizer **authorizer, bool force_new);
   bool ms_verify_authorizer(Connection *con, int peer_type,
 			    int protocol, bufferlist& authorizer_data, bufferlist& authorizer_reply,
-			    bool& isvalid);
+			    bool& isvalid, CryptoKey& session_key);
   bool ms_handle_reset(Connection *con);
   void ms_handle_remote_reset(Connection *con) {}
 
@@ -440,6 +447,7 @@ public:
 
   static int check_features(MonitorStore *store);
 
+  int preinit();
   int init();
   void shutdown();
   void tick();
