@@ -157,7 +157,7 @@ void PGMap::apply_incremental(const Incremental& inc)
     full_ratio = inc.full_ratio;
     ratios_changed = true;
   }
-  if (inc.nearfull_ratio != nearfull_ratio && inc.full_ratio != -1) {
+  if (inc.nearfull_ratio != nearfull_ratio && inc.nearfull_ratio != -1) {
     nearfull_ratio = inc.nearfull_ratio;
     ratios_changed = true;
   }
@@ -309,8 +309,11 @@ void PGMap::stat_pg_sub(const pg_t &pgid, const pg_stat_t &s)
   pg_sum.sub(s);
   if (s.state & PG_STATE_CREATING) {
     creating_pgs.erase(pgid);
-    if (s.acting.size())
+    if (s.acting.size()) {
       creating_pgs_by_osd[s.acting[0]].erase(pgid);
+      if (creating_pgs_by_osd[s.acting[0]].size() == 0)
+        creating_pgs_by_osd.erase(s.acting[0]);
+    }
   }
 }
 
