@@ -116,7 +116,8 @@ void PGMonitor::update_logger()
 
 void PGMonitor::tick() 
 {
-  if (!paxos->is_active()) return;
+  if (!paxos->is_active() ||
+      !mon->is_all_paxos_recovered()) return;
 
   update_from_paxos();
   handle_osd_timeouts();
@@ -544,6 +545,8 @@ struct RetryCheckOSDMap : public Context {
   epoch_t epoch;
   RetryCheckOSDMap(PGMonitor *p, epoch_t e) : pgmon(p), epoch(e) {}
   void finish(int r) {
+    if (r == -ECANCELED)
+      return;
     pgmon->check_osd_map(epoch);
   }
 };
@@ -576,7 +579,7 @@ void PGMonitor::check_osd_map(epoch_t epoch)
        e++) {
     dout(10) << "check_osd_map applying osdmap e" << e << " to pg_map" << dendl;
     bufferlist bl;
-    mon->store->get_bl_sn(bl, "osdmap", e);
+    mon->store->get_bl_sn_safe(bl, "osdmap", e);
     assert(bl.length());
     OSDMap::Incremental inc(bl);
     for (map<int32_t,uint32_t>::iterator p = inc.new_weight.begin();
@@ -772,14 +775,21 @@ void PGMonitor::send_pg_creates()
     vector<int> acting;
     int nrep = mon->osdmon()->osdmap.pg_to_acting_osds(on, acting);
 
-    if (s.acting.size())
+    if (s.acting.size()) {
       pg_map.creating_pgs_by_osd[s.acting[0]].erase(pgid);
+      if (pg_map.creating_pgs_by_osd[s.acting[0]].size() == 0)
+        pg_map.creating_pgs_by_osd.erase(s.acting[0]);
+    }
     s.acting = acting;
 
     // don't send creates for localized pgs
     if (pgid.preferred() >= 0)
       continue;
 
+    // don't send creates for splits
+    if (s.parent_split_bits)
+      continue;
+
     if (nrep) {
       pg_map.creating_pgs_by_osd[acting[0]].insert(pgid);
     } else {
@@ -809,6 +819,7 @@ void PGMonitor::send_pg_creates(int osd, Connection *con)
   map<int, set<pg_t> >::iterator p = pg_map.creating_pgs_by_osd.find(osd);
   if (p == pg_map.creating_pgs_by_osd.end())
     return;
+  assert(p->second.size() > 0);
 
   dout(20) << "send_pg_creates osd." << osd << " pgs " << p->second << dendl;
   MOSDPGCreate *m = new MOSDPGCreate(mon->osdmon()->osdmap.get_epoch());
@@ -988,16 +999,23 @@ bool PGMonitor::preprocess_command(MMonCommand *m)
       r = -EINVAL;
       if (pgid.parse(m->cmd[2].c_str())) {
 	vector<int> up, acting;
-	pg_t mpgid = mon->osdmon()->osdmap.raw_pg_to_pg(pgid);
-	mon->osdmon()->osdmap.pg_to_up_acting_osds(pgid, up, acting);
-	ss << "osdmap e" << mon->osdmon()->osdmap.get_epoch()
-	   << " pg " << pgid << " (" << mpgid << ")"
-	   << " -> up " << up << " acting " << acting;
-	r = 0;
+	if (mon->osdmon()->osdmap.have_pg_pool(pgid.pool())) {
+	  pg_t mpgid = mon->osdmon()->osdmap.raw_pg_to_pg(pgid);
+	  mon->osdmon()->osdmap.pg_to_up_acting_osds(pgid, up, acting);
+	  ss << "osdmap e" << mon->osdmon()->osdmap.get_epoch()
+	    << " pg " << pgid << " (" << mpgid << ")"
+	    << " -> up " << up << " acting " << acting;
+	  r = 0;
+	} else {
+	  r = -ENOENT;
+	  ss << "pg '" << m->cmd[2] << "' does not exist";
+	}
       } else
 	ss << "invalid pgid '" << m->cmd[2] << "'";
     }
-    else if ((m->cmd[1] == "scrub" || m->cmd[1] == "repair") && m->cmd.size() == 3) {
+    else if ((m->cmd[1] == "scrub" ||
+	      m->cmd[1] == "deep-scrub" ||
+	      m->cmd[1] == "repair") && m->cmd.size() == 3) {
       pg_t pgid;
       r = -EINVAL;
       if (pgid.parse(m->cmd[2].c_str())) {
@@ -1008,7 +1026,8 @@ bool PGMonitor::preprocess_command(MMonCommand *m)
 	      vector<pg_t> pgs(1);
 	      pgs[0] = pgid;
 	      mon->try_send_message(new MOSDScrub(mon->monmap->fsid, pgs,
-						  m->cmd[1] == "repair"),
+						  m->cmd[1] == "repair",
+						  m->cmd[1] == "deep-scrub"),
 				    mon->osdmon()->osdmap.get_inst(osd));
 	      ss << "instructing pg " << pgid << " on osd." << osd << " to " << m->cmd[1];
 	      r = 0;
@@ -1219,10 +1238,16 @@ void PGMonitor::get_health(list<pair<health_status_t,string> >& summary,
       note["splitting"] += p->second;
     if (p->first & PG_STATE_RECOVERING)
       note["recovering"] += p->second;
+    if (p->first & PG_STATE_RECOVERY_WAIT)
+      note["recovery_wait"] += p->second;
     if (p->first & PG_STATE_INCOMPLETE)
       note["incomplete"] += p->second;
-    if (p->first & PG_STATE_BACKFILL)
+    if (p->first & PG_STATE_BACKFILL_WAIT)
       note["backfill"] += p->second;
+    if (p->first & PG_STATE_BACKFILL)
+      note["backfilling"] += p->second;
+    if (p->first & PG_STATE_BACKFILL_TOOFULL)
+      note["backfill_toofull"] += p->second;
   }
 
   hash_map<pg_t, pg_stat_t> stuck_pgs;
@@ -1270,14 +1295,24 @@ void PGMonitor::get_health(list<pair<health_status_t,string> >& summary,
 			       PG_STATE_REPAIR |
 			       PG_STATE_SPLITTING |
 			       PG_STATE_RECOVERING |
+			       PG_STATE_RECOVERY_WAIT |
 			       PG_STATE_INCOMPLETE |
-			       PG_STATE_BACKFILL) &&
+			       PG_STATE_BACKFILL_WAIT |
+			       PG_STATE_BACKFILL |
+			       PG_STATE_BACKFILL_TOOFULL) &&
 	    stuck_pgs.count(p->first) == 0) {
 	  ostringstream ss;
 	  ss << "pg " << p->first << " is " << pg_state_string(p->second.state);
 	  ss << ", acting " << p->second.acting;
 	  if (p->second.stats.sum.num_objects_unfound)
 	    ss << ", " << p->second.stats.sum.num_objects_unfound << " unfound";
+	  if (p->second.state & PG_STATE_INCOMPLETE) {
+	    const pg_pool_t *pi = mon->osdmon()->osdmap.get_pg_pool(p->first.pool());
+	    if (pi && pi->min_size > 1) {
+	      ss << " (reducing pool " << mon->osdmon()->osdmap.get_pool_name(p->first.pool())
+		 << " min_size from " << (int)pi->min_size << " may help; search ceph.com/docs for 'incomplete')";
+	    }
+	  }
 	  detail->push_back(make_pair(HEALTH_WARN, ss.str()));
 	}
       }
