@@ -7,12 +7,10 @@
 #include <map>
 #include <string>
 
-#include <string>
 #include <vector>
 #include <tr1/memory>
 #include <boost/scoped_ptr.hpp>
 
-#include "IndexManager.h"
 #include "ObjectMap.h"
 #include "KeyValueDB.h"
 #include "osd/osd_types.h"
@@ -28,10 +26,8 @@
  * @see user_prefix
  * @see sys_prefix
  *
- * - LEAF_PREFIX: Contains mapping from (coll_t,hobject_t)->seq
- * - REVERSE_LEAF_PREFIX: Contains mapping from seq->[(coll_t, hobject_t)]
- *                        @see set_map_header
- *                        @see remove_map_header
+ * - HOBJECT_TO_SEQ: Contains leaf mapping from hobject_t->seq and
+ *                   corresponding omap header
  * - SYS_PREFIX: GLOBAL_STATE_KEY - contains next seq number
  *                                  @see State
  *                                  @see write_state
@@ -45,7 +41,7 @@
  *              : USER_HEADER_KEY - omap header for header->seq
  *              : HEADER_KEY - encoding of header for header->seq
  *
- * For each node (represented by a header, not counting LEAF_PREFIX space), we
+ * For each node (represented by a header), we
  * store three mappings: the key mapping, the complete mapping, and the parent.
  * The complete mapping (COMPLETE_PREFIX space) is key->key.  Each x->y entry in
  * this mapping indicates that the key mapping contains all entries on [x,y).
@@ -59,133 +55,116 @@ class DBObjectMap : public ObjectMap {
 public:
   boost::scoped_ptr<KeyValueDB> db;
 
-  /**
-   * Each header has a unique id.  This value is updated and written *before*
-   * the transaction writing th new header
-   */
-  uint64_t next_seq;
-
   /**
    * Serializes access to next_seq as well as the in_use set
    */
   Mutex header_lock;
   Cond header_cond;
+  Cond map_header_cond;
 
   /**
    * Set of headers currently in use
    */
   set<uint64_t> in_use;
+  set<hobject_t> map_header_in_use;
 
-  DBObjectMap(KeyValueDB *db) : db(db), next_seq(1),
+  DBObjectMap(KeyValueDB *db) : db(db),
 				header_lock("DBOBjectMap")
     {}
 
   int set_keys(
     const hobject_t &hoid,
-    Index index,
-    const map<string, bufferlist> &set
+    const map<string, bufferlist> &set,
+    const SequencerPosition *spos=0
     );
 
   int set_header(
     const hobject_t &hoid,
-    Index index,
-    const bufferlist &bl
+    const bufferlist &bl,
+    const SequencerPosition *spos=0
     );
 
   int get_header(
     const hobject_t &hoid,
-    Index index,
     bufferlist *bl
     );
 
   int clear(
     const hobject_t &hoid,
-    Index index
+    const SequencerPosition *spos=0
     );
 
   int rm_keys(
     const hobject_t &hoid,
-    Index index,
-    const set<string> &to_clear
+    const set<string> &to_clear,
+    const SequencerPosition *spos=0
     );
 
   int get(
     const hobject_t &hoid,
-    Index index,
     bufferlist *header,
     map<string, bufferlist> *out
     );
 
   int get_keys(
     const hobject_t &hoid,
-    Index index,
     set<string> *keys
     );
 
   int get_values(
     const hobject_t &hoid,
-    Index index,
     const set<string> &keys,
     map<string, bufferlist> *out
     );
 
   int check_keys(
     const hobject_t &hoid,
-    Index index,
     const set<string> &keys,
     set<string> *out
     );
 
   int get_xattrs(
     const hobject_t &hoid,
-    Index index,
     const set<string> &to_get,
     map<string, bufferlist> *out
     );
 
   int get_all_xattrs(
     const hobject_t &hoid,
-    Index index,
     set<string> *out
     );
 
   int set_xattrs(
     const hobject_t &hoid,
-    Index index,
-    const map<string, bufferlist> &to_set
+    const map<string, bufferlist> &to_set,
+    const SequencerPosition *spos=0
     );
 
   int remove_xattrs(
     const hobject_t &hoid,
-    Index index,
-    const set<string> &to_remove
+    const set<string> &to_remove,
+    const SequencerPosition *spos=0
     );
 
   int clone(
     const hobject_t &hoid,
-    Index index,
     const hobject_t &target,
-    Index target_index
-    );
-
-  int link(
-    const hobject_t &hoid,
-    Index index,
-    const hobject_t &target,
-    Index target_index
+    const SequencerPosition *spos=0
     );
 
   /// Read initial state from backing store
-  int init();
+  int init(bool upgrade = false);
+
+  /// Upgrade store to current version
+  int upgrade();
 
   /// Consistency check, debug, there must be no parallel writes
   bool check(std::ostream &out);
 
   /// Ensure that all previous operations are durable
-  int sync();
+  int sync(const hobject_t *hoid=0, const SequencerPosition *spos=0);
 
-  ObjectMapIterator get_iterator(const hobject_t &hoid,
-				 Index index);
+  ObjectMapIterator get_iterator(const hobject_t &hoid);
 
   static const string USER_PREFIX;
   static const string XATTR_PREFIX;
@@ -193,24 +172,33 @@ public:
   static const string COMPLETE_PREFIX;
   static const string HEADER_KEY;
   static const string USER_HEADER_KEY;
-  static const string LEAF_PREFIX;
   static const string GLOBAL_STATE_KEY;
+  static const string HOBJECT_TO_SEQ;
+
+  /// Legacy
+  static const string LEAF_PREFIX;
   static const string REVERSE_LEAF_PREFIX;
 
   /// persistent state for store @see generate_header
   struct State {
+    __u8 v;
     uint64_t seq;
-    State() : seq(0) {}
-    State(uint64_t seq) : seq(seq) {}
+    State() : v(0), seq(1) {}
+    State(uint64_t seq) : v(0), seq(seq) {}
 
     void encode(bufferlist &bl) const {
-      ENCODE_START(1, 1, bl);
+      ENCODE_START(2, 1, bl);
+      ::encode(v, bl);
       ::encode(seq, bl);
       ENCODE_FINISH(bl);
     }
 
     void decode(bufferlist::iterator &bl) {
-      DECODE_START(1, bl);
+      DECODE_START(2, bl);
+      if (struct_v >= 2)
+	::decode(v, bl);
+      else
+	v = 0;
       ::decode(seq, bl);
       DECODE_FINISH(bl);
     }
@@ -223,7 +211,7 @@ public:
       o.push_back(new State(0));
       o.push_back(new State(20));
     }
-  };
+  } state;
 
   struct _Header {
     uint64_t seq;
@@ -233,23 +221,28 @@ public:
     coll_t c;
     hobject_t hoid;
 
+    SequencerPosition spos;
+
     void encode(bufferlist &bl) const {
-      ENCODE_START(1, 1, bl);
+      ENCODE_START(2, 1, bl);
       ::encode(seq, bl);
       ::encode(parent, bl);
       ::encode(num_children, bl);
       ::encode(c, bl);
       ::encode(hoid, bl);
+      ::encode(spos, bl);
       ENCODE_FINISH(bl);
     }
 
     void decode(bufferlist::iterator &bl) {
-      DECODE_START(1, bl);
+      DECODE_START(2, bl);
       ::decode(seq, bl);
       ::decode(parent, bl);
       ::decode(num_children, bl);
       ::decode(c, bl);
       ::decode(hoid, bl);
+      if (struct_v >= 2)
+	::decode(spos, bl);
       DECODE_FINISH(bl);
     }
 
@@ -271,13 +264,16 @@ public:
     _Header() : seq(0), parent(0), num_children(1) {}
   };
 
+  /// String munging (public for testing)
+  static string hobject_key(const hobject_t &hoid);
+  static string hobject_key_v0(coll_t c, const hobject_t &hoid);
+  static bool parse_hobject_key_v0(const string &in,
+				   coll_t *c, hobject_t *hoid);
 private:
   /// Implicit lock on Header->seq
   typedef std::tr1::shared_ptr<_Header> Header;
 
-  /// String munging
-  string hobject_key(coll_t c, const hobject_t &hoid);
-  string map_header_key(coll_t c, const hobject_t &hoid);
+  string map_header_key(const hobject_t &hoid);
   string header_key(uint64_t seq);
   string complete_prefix(Header header);
   string user_prefix(Header header);
@@ -369,16 +365,21 @@ private:
   void set_header(Header input, KeyValueDB::Transaction t);
 
   /// Remove leaf node corresponding to hoid in c
-  void remove_map_header(coll_t c, const hobject_t &hoid,
+  void remove_map_header(const hobject_t &hoid,
 			 Header header,
 			 KeyValueDB::Transaction t);
 
   /// Set leaf node for c and hoid to the value of header
-  void set_map_header(coll_t c, const hobject_t &hoid, _Header header,
-		     KeyValueDB::Transaction t);
+  void set_map_header(const hobject_t &hoid, _Header header,
+		      KeyValueDB::Transaction t);
+
+  /// Set leaf node for c and hoid to the value of header
+  bool check_spos(const hobject_t &hoid,
+		  Header header,
+		  const SequencerPosition *spos);
 
   /// Lookup or create header for c hoid
-  Header lookup_create_map_header(coll_t c, const hobject_t &hoid,
+  Header lookup_create_map_header(const hobject_t &hoid,
 				  KeyValueDB::Transaction t);
 
   /**
@@ -386,10 +387,18 @@ private:
    *
    * Has the side effect of syncronously saving the new DBObjectMap state
    */
-  Header generate_new_header(coll_t c, const hobject_t &hoid, Header parent);
+  Header _generate_new_header(const hobject_t &hoid, Header parent);
+  Header generate_new_header(const hobject_t &hoid, Header parent) {
+    Mutex::Locker l(header_lock);
+    return _generate_new_header(hoid, parent);
+  }
 
   /// Lookup leaf header for c hoid
-  Header lookup_map_header(coll_t c, const hobject_t &hoid);
+  Header _lookup_map_header(const hobject_t &hoid);
+  Header lookup_map_header(const hobject_t &hoid) {
+    Mutex::Locker l(header_lock);
+    return _lookup_map_header(hoid);
+  }
 
   /// Lookup header node for input
   Header lookup_parent(Header input);
@@ -414,7 +423,8 @@ private:
 			 KeyValueDB::Transaction t);
 
   /// Writes out State (mainly next_seq)
-  int write_state(bool sync = false);
+  int write_state(KeyValueDB::Transaction _t =
+		  KeyValueDB::Transaction());
 
   /// 0 if the complete set now contains all of key space, < 0 on error, 1 else
   int need_parent(DBObjectMapIterator iter);
@@ -427,6 +437,23 @@ private:
   void _set_header(Header header, const bufferlist &bl,
 		   KeyValueDB::Transaction t);
 
+  /** 
+   * Removes map header lock once Header is out of scope
+   * @see lookup_map_header
+   */
+  class RemoveMapHeaderOnDelete {
+  public:
+    DBObjectMap *db;
+    hobject_t obj;
+    RemoveMapHeaderOnDelete(DBObjectMap *db, const hobject_t &obj) :
+      db(db), obj(obj) {}
+    void operator() (_Header *header) {
+      Mutex::Locker l(db->header_lock);
+      db->map_header_in_use.erase(obj);
+      db->map_header_cond.Signal();
+      delete header;
+    }
+  };
 
   /** 
    * Removes header seq lock once Header is out of scope
@@ -436,9 +463,8 @@ private:
   class RemoveOnDelete {
   public:
     DBObjectMap *db;
-    uint64_t seq;
     RemoveOnDelete(DBObjectMap *db) :
-      db(db), seq(seq) {}
+      db(db) {}
     void operator() (_Header *header) {
       Mutex::Locker l(db->header_lock);
       db->in_use.erase(header->seq);
