@@ -355,6 +355,19 @@ void ObjectStore::Transaction::dump(ceph::Formatter *f)
       }
       break;
 
+    case Transaction::OP_SPLIT_COLLECTION:
+      {
+	coll_t cid(i.get_cid());
+	uint32_t bits(i.get_u32());
+	uint32_t rem(i.get_u32());
+	coll_t dest(i.get_cid());
+	f->dump_string("op_name", "op_split_collection");
+	f->dump_stream("collection") << cid;
+	f->dump_stream("bits") << bits;
+	f->dump_stream("rem") << rem;
+	f->dump_stream("dest") << dest;
+      }
+
     default:
       f->dump_string("op_name", "unknown");
       f->dump_unsigned("op_code", op);
@@ -378,9 +391,9 @@ void ObjectStore::Transaction::generate_test_instances(list<ObjectStore::Transac
   t = new Transaction;
   coll_t c("foocoll");
   coll_t c2("foocoll2");
-  hobject_t o1("obj", "", 123, 456);
-  hobject_t o2("obj2", "", 123, 456);
-  hobject_t o3("obj3", "", 123, 456);
+  hobject_t o1("obj", "", 123, 456, -1);
+  hobject_t o2("obj2", "", 123, 456, -1);
+  hobject_t o3("obj3", "", 123, 456, -1);
   t->touch(c, o1);
   bufferlist bl;
   bl.append("some data");
