@@ -49,10 +49,10 @@ using namespace std;
 #include <ext/hash_set>
 using namespace __gnu_cxx;
 
-#include "OpRequest.h"
 #include "common/shared_cache.hpp"
 #include "common/simple_cache.hpp"
 #include "common/sharedptr_registry.hpp"
+#include "common/PrioritizedQueue.h"
 
 #define CEPH_OSD_PROTOCOL    10 /* cluster internal */
 
@@ -133,6 +133,7 @@ class AuthAuthorizeHandlerRegistry;
 
 class OpsFlightSocketHook;
 class HistoricOpsSocketHook;
+struct C_CompleteSplits;
 
 extern const coll_t meta_coll;
 
@@ -168,11 +169,13 @@ public:
   ObjectStore *&store;
   LogClient &clog;
   PGRecoveryStats &pg_recovery_stats;
+private:
   Messenger *&cluster_messenger;
   Messenger *&client_messenger;
+public:
   PerfCounters *&logger;
   MonClient   *&monc;
-  ThreadPool::WorkQueue<PG> &op_wq;
+  ThreadPool::WorkQueueVal<pair<PGRef, OpRequestRef>, PGRef> &op_wq;
   ThreadPool::BatchWorkQueue<PG> &peering_wq;
   ThreadPool::WorkQueue<PG> &recovery_wq;
   ThreadPool::WorkQueue<PG> &snap_trim_wq;
@@ -181,8 +184,10 @@ public:
   ThreadPool::WorkQueue<MOSDRepScrub> &rep_scrub_wq;
   ClassHandler  *&class_handler;
 
+  void dequeue_pg(PG *pg, list<OpRequestRef> *dequeued);
+
   // -- superblock --
-  Mutex publish_lock;
+  Mutex publish_lock, pre_publish_lock;
   OSDSuperblock superblock;
   OSDSuperblock get_superblock() {
     Mutex::Locker l(publish_lock);
@@ -192,6 +197,9 @@ public:
     Mutex::Locker l(publish_lock);
     superblock = block;
   }
+
+  int get_nodeid() const { return whoami; }
+
   OSDMapRef osdmap;
   OSDMapRef get_osdmap() {
     Mutex::Locker l(publish_lock);
@@ -202,7 +210,37 @@ public:
     osdmap = map;
   }
 
-  int get_nodeid() const { return whoami; }
+  /*
+   * osdmap - current published amp
+   * next_osdmap - pre_published map that is about to be published.
+   *
+   * We use the next_osdmap to send messages and initiate connections,
+   * but only if the target is the same instance as the one in the map
+   * epoch the current user is working from (i.e., the result is
+   * equivalent to what is in next_osdmap).
+   *
+   * This allows the helpers to start ignoring osds that are about to
+   * go down, and let OSD::handle_osd_map()/note_down_osd() mark them
+   * down, without worrying about reopening connections from threads
+   * working from old maps.
+   */
+  OSDMapRef next_osdmap;
+  void pre_publish_map(OSDMapRef map) {
+    Mutex::Locker l(pre_publish_lock);
+    next_osdmap = map;
+  }
+  ConnectionRef get_con_osd_cluster(int peer, epoch_t from_epoch);
+  ConnectionRef get_con_osd_hb(int peer, epoch_t from_epoch);
+  void send_message_osd_cluster(int peer, Message *m, epoch_t from_epoch);
+  void send_message_osd_cluster(Message *m, Connection *con) {
+    cluster_messenger->send_message(m, con);
+  }
+  void send_message_osd_client(Message *m, Connection *con) {
+    client_messenger->send_message(m, con);
+  }
+  entity_name_t get_cluster_msgr_name() {
+    return cluster_messenger->get_myname();
+  }
 
   // -- scrub scheduling --
   Mutex sched_scrub_lock;
@@ -235,6 +273,7 @@ public:
   }
 
   bool inc_scrubs_pending();
+  void inc_scrubs_active(bool reserved);
   void dec_scrubs_pending();
   void dec_scrubs_active();
 
@@ -279,7 +318,6 @@ public:
   void send_pg_temp();
 
   void queue_for_peering(PG *pg);
-  void queue_for_op(PG *pg);
   bool queue_for_recovery(PG *pg);
   bool queue_for_snap_trim(PG *pg) {
     return snap_trim_wq.queue(pg);
@@ -321,7 +359,7 @@ public:
   void _add_map_inc_bl(epoch_t e, bufferlist& bl);
   bool get_inc_map_bl(epoch_t e, bufferlist& bl);
 
-  void clear_map_bl_cache_pins();
+  void clear_map_bl_cache_pins(epoch_t e);
 
   void need_heartbeat_peer_update();
 
@@ -331,6 +369,19 @@ public:
   void init();
   void shutdown();
 
+  // split
+  Mutex in_progress_split_lock;
+  set<pg_t> in_progress_splits;
+  void _start_split(const set<pg_t> &pgs);
+  void start_split(const set<pg_t> &pgs) {
+    Mutex::Locker l(in_progress_split_lock);
+    return _start_split(pgs);
+  }
+  void complete_split(const set<pg_t> &pgs);
+  bool splitting(pg_t pgid);
+  void expand_pg_num(OSDMapRef old_map,
+		     OSDMapRef new_map);
+
   OSDService(OSD *osd);
 };
 class OSD : public Dispatcher {
@@ -416,9 +467,10 @@ private:
 
   // -- state --
 public:
-  static const int STATE_BOOTING = 1;
-  static const int STATE_ACTIVE = 2;
-  static const int STATE_STOPPING = 3;
+  static const int STATE_INITIALIZING = 1;
+  static const int STATE_BOOTING = 2;
+  static const int STATE_ACTIVE = 3;
+  static const int STATE_STOPPING = 4;
 
 private:
   int state;
@@ -427,6 +479,7 @@ private:
   epoch_t bind_epoch;  // epoch we last did a bind to new ip:ports
 
 public:
+  bool is_initializing() { return state == STATE_INITIALIZING; }
   bool is_booting() { return state == STATE_BOOTING; }
   bool is_active() { return state == STATE_ACTIVE; }
   bool is_stopping() { return state == STATE_STOPPING; }
@@ -449,24 +502,15 @@ public:
     epoch_t last_sent_epoch;
     Connection *con;
     std::map<void *, pg_t> watches;
-    std::map<void *, entity_name_t> notifs;
 
     Session() : auid(-1), last_sent_epoch(0), con(0) {}
-    void add_notif(void *n, entity_name_t& name) {
-      notifs[n] = name;
-    }
-    void del_notif(void *n) {
-      std::map<void *, entity_name_t>::iterator iter = notifs.find(n);
-      if (iter != notifs.end())
-        notifs.erase(iter);
-    }
   };
 
 private:
   // -- heartbeat --
   /// information about a heartbeat peer
   struct HeartbeatInfo {
-    entity_inst_t inst; ///< peer
+    int peer;           ///< peer
     Connection *con;    ///< peer connection
     utime_t first_tx;   ///< time we sent our first ping request
     utime_t last_tx;    ///< last time we sent a ping request
@@ -518,6 +562,12 @@ public:
       return osd->heartbeat_reset(con);
     }
     void ms_handle_remote_reset(Connection *con) {}
+    bool ms_verify_authorizer(Connection *con, int peer_type,
+			      int protocol, bufferlist& authorizer_data, bufferlist& authorizer_reply,
+			      bool& isvalid, CryptoKey& session_key) {
+      isvalid = true;
+      return true;
+    }
   public:
     OSD *osd;
     HeartbeatDispatcher(OSD *o) 
@@ -559,48 +609,65 @@ private:
   }
   friend class OpsFlightSocketHook;
   friend class HistoricOpsSocketHook;
+  friend class C_CompleteSplits;
   OpsFlightSocketHook *admin_ops_hook;
   HistoricOpsSocketHook *historic_ops_hook;
 
   // -- op queue --
-  list<PG*> op_queue;
-  int op_queue_len;
 
-  struct OpWQ : public ThreadPool::WorkQueue<PG> {
+  struct OpWQ: public ThreadPool::WorkQueueVal<pair<PGRef, OpRequestRef>,
+					       PGRef > {
+    Mutex qlock;
+    map<PG*, list<OpRequestRef> > pg_for_processing;
     OSD *osd;
+    PrioritizedQueue<pair<PGRef, OpRequestRef>, entity_inst_t > pqueue;
     OpWQ(OSD *o, time_t ti, ThreadPool *tp)
-      : ThreadPool::WorkQueue<PG>("OSD::OpWQ", ti, ti*10, tp), osd(o) {}
+      : ThreadPool::WorkQueueVal<pair<PGRef, OpRequestRef>, PGRef >(
+	"OSD::OpWQ", ti, ti*10, tp),
+	qlock("OpWQ::qlock"),
+	osd(o) {}
 
-    bool _enqueue(PG *pg);
-    void _dequeue(PG *pg) {
-      for (list<PG*>::iterator i = osd->op_queue.begin();
-	   i != osd->op_queue.end();
-	   ) {
-	if (*i == pg) {
-	  osd->op_queue.erase(i++);
-	  pg->put();
-	} else {
-	  ++i;
+    void _enqueue_front(pair<PGRef, OpRequestRef> item);
+    void _enqueue(pair<PGRef, OpRequestRef> item);
+    PGRef _dequeue();
+
+    struct Pred {
+      PG *pg;
+      Pred(PG *pg) : pg(pg) {}
+      bool operator()(const pair<PGRef, OpRequestRef> &op) {
+	return op.first == pg;
+      }
+    };
+    void dequeue(PG *pg, list<OpRequestRef> *dequeued = 0) {
+      lock();
+      if (!dequeued) {
+	pqueue.remove_by_filter(Pred(pg));
+	pg_for_processing.erase(pg);
+      } else {
+	list<pair<PGRef, OpRequestRef> > _dequeued;
+	pqueue.remove_by_filter(Pred(pg), &_dequeued);
+	for (list<pair<PGRef, OpRequestRef> >::iterator i = _dequeued.begin();
+	     i != _dequeued.end();
+	     ++i) {
+	  dequeued->push_back(i->second);
+	}
+	if (pg_for_processing.count(pg)) {
+	  dequeued->splice(
+	    dequeued->begin(),
+	    pg_for_processing[pg]);
+	  pg_for_processing.erase(pg);
 	}
       }
+      unlock();
     }
     bool _empty() {
-      return osd->op_queue.empty();
-    }
-    PG *_dequeue();
-    void _process(PG *pg) {
-      osd->dequeue_op(pg);
-    }
-    void _clear() {
-      assert(osd->op_queue.empty());
+      return pqueue.empty();
     }
+    void _process(PGRef pg);
   } op_wq;
 
   void enqueue_op(PG *pg, OpRequestRef op);
-  void dequeue_op(PG *pg);
-  static void static_dequeueop(OSD *o, PG *pg) {
-    o->dequeue_op(pg);
-  };
+  void dequeue_op(PGRef pg, OpRequestRef op);
 
   // -- peering queue --
   struct PeeringWQ : public ThreadPool::BatchWorkQueue<PG> {
@@ -693,7 +760,7 @@ private:
 
   bool _share_map_incoming(const entity_inst_t& inst, epoch_t epoch,
 			   Session *session = 0);
-  void _share_map_outgoing(const entity_inst_t& inst,
+  void _share_map_outgoing(int peer, Connection *con,
 			   OSDMapRef map = OSDMapRef());
 
   void wait_for_new_map(OpRequestRef op);
@@ -701,7 +768,9 @@ private:
   void note_down_osd(int osd);
   void note_up_osd(int osd);
   
-  void advance_pg(epoch_t advance_to, PG *pg, PG::RecoveryCtx *rctx);
+  void advance_pg(
+    epoch_t advance_to, PG *pg, PG::RecoveryCtx *rctx,
+    set<boost::intrusive_ptr<PG> > *split_pgs);
   void advance_map(ObjectStore::Transaction& t, C_Contexts *tfin);
   void activate_map();
 
@@ -730,18 +799,18 @@ private:
   bool get_inc_map_bl(epoch_t e, bufferlist& bl) {
     return service.get_inc_map_bl(e, bl);
   }
-  void clear_map_bl_cache_pins() {
-    service.clear_map_bl_cache_pins();
-  }
 
   MOSDMap *build_incremental_map_msg(epoch_t from, epoch_t to);
   void send_incremental_map(epoch_t since, const entity_inst_t& inst, bool lazy=false);
+  void send_incremental_map(epoch_t since, Connection *con);
   void send_map(MOSDMap *m, const entity_inst_t& inst, bool lazy);
+  void send_map(MOSDMap *m, Connection *con);
 
 protected:
   // -- placement groups --
   hash_map<pg_t, PG*> pg_map;
   map<pg_t, list<OpRequestRef> > waiting_for_pg;
+  map<pg_t, list<PG::CephPeeringEvtRef> > peering_wait_for_split;
   PGRecoveryStats pg_recovery_stats;
 
   PGPool _get_pool(int id, OSDMapRef createmap);
@@ -763,11 +832,14 @@ protected:
   PG   *_lookup_qlock_pg(pg_t pgid);
 
   PG *lookup_lock_raw_pg(pg_t pgid);
+  PG* _make_pg(OSDMapRef createmap, pg_t pgid);
+  void add_newly_split_pg(PG *pg,
+			  PG::RecoveryCtx *rctx);
 
   PG *get_or_create_pg(const pg_info_t& info,
-		       pg_interval_map_t& pi,
-		       epoch_t epoch, int from, int& pcreated,
-		       bool primary);
+                       pg_interval_map_t& pi,
+                       epoch_t epoch, int from, int& pcreated,
+                       bool primary);
   
   void load_pgs();
   void build_past_intervals_parallel();
@@ -797,7 +869,6 @@ protected:
     vector<int> acting;
     set<int> prior;
     pg_t parent;
-    int split_bits;
   };
   hash_map<pg_t, create_pg_info> creating_pgs;
   double debug_drop_pg_create_probability;
@@ -809,7 +880,12 @@ protected:
 
   void do_split(PG *parent, set<pg_t>& children, ObjectStore::Transaction &t, C_Contexts *tfin);
   void split_pg(PG *parent, map<pg_t,PG*>& children, ObjectStore::Transaction &t);
-
+  void split_pgs(
+    PG *parent,
+    const set<pg_t> &childpgids, set<boost::intrusive_ptr<PG> > *out_pgs,
+    OSDMapRef curmap,
+    OSDMapRef nextmap,
+    PG::RecoveryCtx *rctx);
 
   // == monitor interaction ==
   utime_t last_mon_report;
@@ -917,6 +993,7 @@ protected:
 
   void handle_pg_backfill(OpRequestRef op);
   void handle_pg_backfill_reserve(OpRequestRef op);
+  void handle_pg_recovery_reserve(OpRequestRef op);
 
   void handle_pg_remove(OpRequestRef op);
   void _remove_pg(PG *pg);
@@ -1278,7 +1355,7 @@ protected:
   bool ms_get_authorizer(int dest_type, AuthAuthorizer **authorizer, bool force_new);
   bool ms_verify_authorizer(Connection *con, int peer_type,
 			    int protocol, bufferlist& authorizer, bufferlist& authorizer_reply,
-			    bool& isvalid);
+			    bool& isvalid, CryptoKey& session_key);
   void ms_handle_connect(Connection *con);
   bool ms_handle_reset(Connection *con);
   void ms_handle_remote_reset(Connection *con) {}
