@@ -143,7 +143,14 @@ inline ostream& operator<<(ostream& out, const object_locator_t& loc)
 }
 
 
-
+// Internal OSD op flags - set by the OSD based on the op types
+enum {
+  CEPH_OSD_RMW_FLAG_READ        = (1 << 1),
+  CEPH_OSD_RMW_FLAG_WRITE       = (1 << 2),
+  CEPH_OSD_RMW_FLAG_CLASS_READ  = (1 << 3),
+  CEPH_OSD_RMW_FLAG_CLASS_WRITE = (1 << 4),
+  CEPH_OSD_RMW_FLAG_PGOP        = (1 << 5),
+};
 
 
 // pg stuff
@@ -228,11 +235,19 @@ struct pg_t {
     m_preferred = osd;
   }
 
+  pg_t get_parent() const;
+
   int print(char *o, int maxlen) const;
   bool parse(const char *s);
 
   bool is_split(unsigned old_pg_num, unsigned new_pg_num, set<pg_t> *pchildren) const;
 
+  /**
+   * Returns b such that for all object o:
+   *   ~((~0)<<b) & o.hash) == 0 iff o is in the pg for *this
+   */
+  unsigned get_split_bits(unsigned pg_num) const;
+
   void encode(bufferlist& bl) const {
     __u8 v = 1;
     ::encode(v, bl);
@@ -567,6 +582,7 @@ inline ostream& operator<<(ostream& out, const osd_stat_t& s) {
 #define PG_STATE_DEEP_SCRUB   (1<<19) // deep scrub: check CRC32 on files
 #define PG_STATE_BACKFILL  (1<<20) // [active] backfilling pg content
 #define PG_STATE_BACKFILL_TOOFULL (1<<21) // backfill can't proceed: too full
+#define PG_STATE_RECOVERY_WAIT (1<<22) // waiting for recovery reservations
 
 std::string pg_state_string(int state);
 
@@ -582,11 +598,11 @@ struct pool_snap_info_t {
   string name;
 
   void dump(Formatter *f) const;
-  void encode(bufferlist& bl) const;
+  void encode(bufferlist& bl, uint64_t features) const;
   void decode(bufferlist::iterator& bl);
   static void generate_test_instances(list<pool_snap_info_t*>& o);
 };
-WRITE_CLASS_ENCODER(pool_snap_info_t)
+WRITE_CLASS_ENCODER_FEATURES(pool_snap_info_t)
 
 inline ostream& operator<<(ostream& out, const pool_snap_info_t& si) {
   return out << si.snapid << '(' << si.name << ' ' << si.stamp << ')';
@@ -615,7 +631,7 @@ struct pg_pool_t {
 
   uint64_t flags;           /// FLAG_* 
   __u8 type;                /// TYPE_*
-  __u8 size;                /// number of osds in each pg
+  __u8 size, min_size;      /// number of osds in each pg
   __u8 crush_ruleset;       /// crush placement rule set
   __u8 object_hash;         /// hash mapping object name to ps
 private:
@@ -644,7 +660,8 @@ public:
   int pg_num_mask, pgp_num_mask;
 
   pg_pool_t()
-    : flags(0), type(0), size(0), crush_ruleset(0), object_hash(0),
+    : flags(0), type(0), size(0), min_size(0),
+      crush_ruleset(0), object_hash(0),
       pg_num(0), pgp_num(0),
       last_change(0),
       snap_seq(0), snap_epoch(0),
@@ -870,6 +887,7 @@ struct pg_stat_t {
   utime_t last_deep_scrub_stamp;
 
   object_stat_collection_t stats;
+  bool stats_invalid;
 
   int64_t log_size;
   int64_t ondisk_log_size;    // >= active_log_size
@@ -881,6 +899,7 @@ struct pg_stat_t {
     : state(0),
       created(0), last_epoch_clean(0),
       parent_split_bits(0), 
+      stats_invalid(false),
       log_size(0), ondisk_log_size(0),
       mapping_epoch(0)
   { }
@@ -932,11 +951,11 @@ struct pool_stat_t {
   }
 
   void dump(Formatter *f) const;
-  void encode(bufferlist &bl) const;
+  void encode(bufferlist &bl, uint64_t features) const;
   void decode(bufferlist::iterator &bl);
   static void generate_test_instances(list<pool_stat_t*>& o);
 };
-WRITE_CLASS_ENCODER(pool_stat_t)
+WRITE_CLASS_ENCODER_FEATURES(pool_stat_t)
 
 
 /**
@@ -1030,6 +1049,7 @@ struct pg_info_t {
   pg_t pgid;
   eversion_t last_update;    // last object version applied to store.
   eversion_t last_complete;  // last version pg was complete through.
+  epoch_t last_epoch_started;// last epoch at which this pg started on this osd
   
   eversion_t log_tail;     // oldest log entry.
 
@@ -1042,11 +1062,11 @@ struct pg_info_t {
   pg_history_t history;
 
   pg_info_t()
-    : last_backfill(hobject_t::get_max())
+    : last_epoch_started(0), last_backfill(hobject_t::get_max())
   { }
   pg_info_t(pg_t p)
     : pgid(p),
-      last_backfill(hobject_t::get_max())
+      last_epoch_started(0), last_backfill(hobject_t::get_max())
   { }
   
   bool is_empty() const { return last_update.version == 0; }
@@ -1077,6 +1097,7 @@ inline ostream& operator<<(ostream& out, const pg_info_t& pgi)
       out << " lb " << pgi.last_backfill;
   }
   //out << " c " << pgi.epoch_created;
+  out << " local-les=" << pgi.last_epoch_started;
   out << " n=" << pgi.stats.stats.sum.num_objects;
   out << " " << pgi.history
       << ")";
@@ -1132,6 +1153,8 @@ struct pg_interval_t {
     epoch_t last_epoch_clean,                   ///< [in] current
     std::tr1::shared_ptr<const OSDMap> osdmap,  ///< [in] current map
     std::tr1::shared_ptr<const OSDMap> lastmap, ///< [in] last map
+    int64_t poolid,                             ///< [in] pool for pg
+    pg_t pgid,                                  ///< [in] pgid for pg
     map<epoch_t, pg_interval_t> *past_intervals,///< [out] intervals
     ostream *out = 0                            ///< [out] debug ostream
     );
@@ -1241,7 +1264,7 @@ struct pg_log_entry_t {
 
   __s32      op;
   hobject_t  soid;
-  eversion_t version, prior_version;
+  eversion_t version, prior_version, reverting_to;
   osd_reqid_t reqid;  // caller+tid to uniquely identify request
   utime_t     mtime;  // this is the _user_ mtime, mind you
   bufferlist snaps;   // only for clone entries
@@ -1437,6 +1460,7 @@ struct pg_missing_t {
   void rm(const std::map<hobject_t, pg_missing_t::item>::iterator &m);
   void got(const hobject_t& oid, eversion_t v);
   void got(const std::map<hobject_t, pg_missing_t::item>::iterator &m);
+  void split_into(pg_t child_pgid, unsigned split_bits, pg_missing_t *omissing);
 
   void clear() {
     missing.clear();
@@ -1540,22 +1564,24 @@ ostream& operator<<(ostream& out, const osd_peer_stat_t &stat);
 class ObjectExtent {
  public:
   object_t    oid;       // object id
+  uint64_t    objectno;
   uint64_t    offset;    // in object
   uint64_t    length;    // in object
 
   object_locator_t oloc;   // object locator (pool etc)
 
-  map<uint64_t, uint64_t>  buffer_extents;  // off -> len.  extents in buffer being mapped (may be fragmented bc of striping!)
+  vector<pair<uint64_t,uint64_t> >  buffer_extents;  // off -> len.  extents in buffer being mapped (may be fragmented bc of striping!)
   
-  ObjectExtent() : offset(0), length(0) {}
-  ObjectExtent(object_t o, uint64_t off=0, uint64_t l=0) : oid(o), offset(off), length(l) { }
+  ObjectExtent() : objectno(0), offset(0), length(0) {}
+  ObjectExtent(object_t o, uint64_t ono, uint64_t off, uint64_t l) : oid(o), objectno(ono), offset(off), length(l) { }
 };
 
 inline ostream& operator<<(ostream& out, const ObjectExtent &ex)
 {
   return out << "extent(" 
-             << ex.oid << " in " << ex.oloc
+             << ex.oid << " (" << ex.objectno << ") in " << ex.oloc
              << " " << ex.offset << "~" << ex.length
+	     << " -> " << ex.buffer_extents
              << ")";
 }
 
