@@ -18,8 +18,6 @@
 #include "rados_sync.h"
 using namespace librados;
 
-#include "common/obj_bencher.h"
-
 #include "common/config.h"
 #include "common/ceph_argparse.h"
 #include "global/global_init.h"
@@ -42,8 +40,6 @@ using namespace librados;
 #include <climits>
 #include <locale>
 
-#include "common/errno.h"
-
 #include "cls/lock/cls_lock_client.h"
 
 int rados_tool_sync(const std::map < std::string, std::string > &opts,
@@ -327,49 +323,51 @@ static int do_put(IoCtx& io_ctx, const char *objname, const char *infile, int op
   if (check_stdio && strcmp(infile, "-") == 0)
     stdio = true;
 
-  if (stdio) {
-    char buf[256];
-    while(!cin.eof()) {
-      cin.getline(buf, 256);
-      indata.append(buf);
-      indata.append('\n');
-    }
-  } else {
-    int ret, fd = open(infile, O_RDONLY);
-    if (fd < 0) {
-      char buf[80];
-      cerr << "error reading input file " << infile << ": " << strerror_r(errno, buf, sizeof(buf)) << std::endl;
-      return 1;
-    }
-    char *buf = new char[op_size];
-    int count = op_size;
-    uint64_t offset = 0;
-    while (count == op_size) {
-      count = read(fd, buf, op_size);
-      if (count == 0) {
-        if (!offset) {
-          int ret = io_ctx.create(oid, true);
-          if (ret < 0)
-            cerr << "WARNING: could not create object: " << oid << std::endl;
-        }
-        continue;
-      }
-      indata.append(buf, count);
-      if (offset == 0)
-	ret = io_ctx.write_full(oid, indata);
-      else
-	ret = io_ctx.write(oid, indata, count, offset);
-      indata.clear();
-
-      if (ret < 0) {
-        close(fd);
-        return ret;
-      }
-      offset += count;
-    }
-    close(fd);
+  int ret;
+  int fd = 0;
+  if (!stdio)
+    fd = open(infile, O_RDONLY);
+  if (fd < 0) {
+    char buf[80];
+    cerr << "error reading input file " << infile << ": " << strerror_r(errno, buf, sizeof(buf)) << std::endl;
+    return 1;
   }
-  return 0;
+  char *buf = new char[op_size];
+  int count = op_size;
+  uint64_t offset = 0;
+  while (count != 0) {
+    count = read(fd, buf, op_size);
+    if (count < 0) {
+      ret = -errno;
+      cerr << "error reading input file " << infile << ": " << cpp_strerror(ret) << std::endl;
+      goto out;
+    }
+    if (count == 0) {
+      if (!offset) {
+	ret = io_ctx.create(oid, true);
+	if (ret < 0) {
+	  cerr << "WARNING: could not create object: " << oid << std::endl;
+	  goto out;
+	}
+      }
+      continue;
+    }
+    indata.append(buf, count);
+    if (offset == 0)
+      ret = io_ctx.write_full(oid, indata);
+    else
+      ret = io_ctx.write(oid, indata, count, offset);
+    indata.clear();
+
+    if (ret < 0) {
+      goto out;
+    }
+    offset += count;
+  }
+  ret = 0;
+ out:
+  TEMP_FAILURE_RETRY(close(fd));
+  return ret;
 }
 
 class RadosWatchCtx : public librados::WatchCtx {
@@ -414,14 +412,6 @@ struct obj_info {
   size_t len;
 };
 
-uint64_t get_random(uint64_t min_val, uint64_t max_val)
-{
-  uint64_t r;
-  get_random_bytes((char *)&r, sizeof(r));
-  r = min_val + r % (max_val - min_val + 1);
-  return r;
-}
-
 class LoadGen {
   size_t total_sent;
   size_t total_completed;
