@@ -37,20 +37,24 @@ namespace ceph {
 using ceph::crypto::MD5;
 
 
-#define RGW_ROOT_BUCKET ".rgw"
-
-#define RGW_CONTROL_BUCKET ".rgw.control"
-
 #define RGW_ATTR_PREFIX  "user.rgw."
 
+#define RGW_AMZ_META_PREFIX "x-amz-meta-"
+
 #define RGW_ATTR_ACL		RGW_ATTR_PREFIX "acl"
 #define RGW_ATTR_ETAG    	RGW_ATTR_PREFIX "etag"
 #define RGW_ATTR_BUCKETS	RGW_ATTR_PREFIX "buckets"
-#define RGW_ATTR_META_PREFIX	RGW_ATTR_PREFIX "x-amz-meta-"
+#define RGW_ATTR_META_PREFIX	RGW_ATTR_PREFIX RGW_AMZ_META_PREFIX
 #define RGW_ATTR_CONTENT_TYPE	RGW_ATTR_PREFIX "content_type"
+#define RGW_ATTR_CACHE_CONTROL	RGW_ATTR_PREFIX "cache_control"
+#define RGW_ATTR_CONTENT_DISP	RGW_ATTR_PREFIX "content_disposition"
+#define RGW_ATTR_CONTENT_ENC	RGW_ATTR_PREFIX "content_encoding"
+#define RGW_ATTR_CONTENT_LANG	RGW_ATTR_PREFIX "content_language"
+#define RGW_ATTR_EXPIRES	RGW_ATTR_PREFIX "expires"
 #define RGW_ATTR_ID_TAG    	RGW_ATTR_PREFIX "idtag"
 #define RGW_ATTR_SHADOW_OBJ    	RGW_ATTR_PREFIX "shadow_name"
 #define RGW_ATTR_MANIFEST    	RGW_ATTR_PREFIX "manifest"
+#define RGW_ATTR_USER_MANIFEST  RGW_ATTR_PREFIX "user_manifest"
 
 #define RGW_BUCKETS_OBJ_PREFIX ".buckets"
 
@@ -58,44 +62,26 @@ using ceph::crypto::MD5;
 #define RGW_MAX_PENDING_CHUNKS  16
 #define RGW_MAX_PUT_SIZE        (5ULL*1024*1024*1024)
 
+#define RGW_FORMAT_PLAIN        0
 #define RGW_FORMAT_XML          1
 #define RGW_FORMAT_JSON         2
 
+#define RGW_CAP_READ            0x1
+#define RGW_CAP_WRITE           0x2
+#define RGW_CAP_ALL             (RGW_CAP_READ | RGW_CAP_WRITE)
+
 #define RGW_REST_SWIFT          0x1
 #define RGW_REST_SWIFT_AUTH     0x2
 
 #define RGW_SUSPENDED_USER_AUID (uint64_t)-2
 
-#define CGI_PRINTF(state, format, ...) do { \
-   int __ret = FCGX_FPrintF(state->fcgx->out, format, __VA_ARGS__); \
-   if (state->header_ended) \
-     state->bytes_sent += __ret; \
-   int l = 32, n; \
-   while (1) { \
-     char __buf[l]; \
-     n = snprintf(__buf, sizeof(__buf), format, __VA_ARGS__); \
-     if (n != l) \
-       dout(10) << "--> " << __buf << dendl; \
-       break; \
-     l *= 2; \
-   } \
-} while (0)
-
-#define CGI_PutStr(state, buf, len) do { \
-  FCGX_PutStr(buf, len, state->fcgx->out); \
-  if (state->header_ended) \
-    state->bytes_sent += len; \
-} while (0)
-
-#define CGI_GetStr(state, buf, buf_len, olen) do { \
-  olen = FCGX_GetStr(buf, buf_len, state->fcgx->in); \
-  state->bytes_received += olen; \
-} while (0)
+#define RGW_DEFAULT_MAX_BUCKETS 1000
 
 #define STATUS_CREATED           1900
 #define STATUS_ACCEPTED          1901
 #define STATUS_NO_CONTENT        1902
 #define STATUS_PARTIAL_CONTENT   1903
+#define STATUS_REDIRECT          1904
 
 #define ERR_INVALID_BUCKET_NAME  2000
 #define ERR_INVALID_OBJECT_NAME  2001
@@ -117,6 +103,9 @@ using ceph::crypto::MD5;
 #define ERR_INVALID_UTF8         2017
 #define ERR_UNPROCESSABLE_ENTITY 2018
 #define ERR_TOO_LARGE            2019
+#define ERR_TOO_MANY_BUCKETS     2020
+#define ERR_INVALID_REQUEST      2021
+#define ERR_TOO_SMALL            2022
 #define ERR_USER_SUSPENDED       2100
 #define ERR_INTERNAL_ERROR       2200
 
@@ -149,6 +138,9 @@ enum {
   l_rgw_cache_hit,
   l_rgw_cache_miss,
 
+  l_rgw_keystone_token_cache_hit,
+  l_rgw_keystone_token_cache_miss,
+
   l_rgw_last,
 };
 
@@ -206,16 +198,22 @@ class XMLArgs
   string str, empty_str;
   map<string, string> val_map;
   map<string, string> sub_resources;
+
+  bool has_resp_modifier;
  public:
-  XMLArgs() {}
-  XMLArgs(string s) : str(s) {}
+  XMLArgs() : has_resp_modifier(false) {}
   /** Set the arguments; as received */
-  void set(string s) { val_map.clear(); sub_resources.clear(); str = s; }
+  void set(string s) {
+    has_resp_modifier = false;
+    val_map.clear();
+    sub_resources.clear();
+    str = s;
+  }
   /** parse the received arguments */
   int parse();
   /** Get the value for a specific argument parameter */
-  string& get(string& name);
-  string& get(const char *name);
+  string& get(const string& name, bool *exists = NULL);
+  string& get(const char *name, bool *exists = NULL);
   /** see if a parameter is contained in this XMLArgs */
   bool exists(const char *name) {
     map<string, string>::iterator iter = val_map.find(name);
@@ -226,6 +224,10 @@ class XMLArgs
     return (iter != sub_resources.end());
   }
   map<string, string>& get_sub_resources() { return sub_resources; }
+
+  bool has_response_modifier() {
+    return has_resp_modifier;
+  }
 };
 
 class RGWConf;
@@ -250,9 +252,10 @@ protected:
   void init(CephContext *cct, RGWEnv * env);
 public:
   RGWConf() :
-    should_log(1) {}
+    enable_ops_log(1), enable_usage_log(1) {}
 
-  int should_log;
+  int enable_ops_log;
+  int enable_usage_log;
 };
 
 enum http_op {
@@ -316,6 +319,33 @@ struct RGWSubUser {
 };
 WRITE_CLASS_ENCODER(RGWSubUser);
 
+class RGWUserCaps
+{
+  map<string, uint32_t> caps;
+
+  int get_cap(const string& cap, string& type, uint32_t *perm);
+  int parse_cap_perm(const string& str, uint32_t *perm);
+  int add_cap(const string& cap);
+  int remove_cap(const string& cap);
+public:
+  int add_from_string(const string& str);
+  int remove_from_string(const string& str);
+
+  void encode(bufferlist& bl) const {
+     ENCODE_START(1, 1, bl);
+     ::encode(caps, bl);
+     ENCODE_FINISH(bl);
+  }
+  void decode(bufferlist::iterator& bl) {
+     DECODE_START(1, bl);
+     ::decode(caps, bl);
+     DECODE_FINISH(bl);
+  }
+  int check_cap(const string& cap, uint32_t perm);
+  void dump(Formatter *f) const;
+};
+WRITE_CLASS_ENCODER(RGWUserCaps);
+
 
 struct RGWUserInfo
 {
@@ -327,11 +357,13 @@ struct RGWUserInfo
   map<string, RGWAccessKey> swift_keys;
   map<string, RGWSubUser> subusers;
   __u8 suspended;
+  uint32_t max_buckets;
+  RGWUserCaps caps;
 
-  RGWUserInfo() : auid(0), suspended(0) {}
+  RGWUserInfo() : auid(0), suspended(0), max_buckets(RGW_DEFAULT_MAX_BUCKETS) {}
 
   void encode(bufferlist& bl) const {
-     ENCODE_START(9, 9, bl);
+     ENCODE_START(11, 9, bl);
      ::encode(auid, bl);
      string access_key;
      string secret_key;
@@ -360,10 +392,12 @@ struct RGWUserInfo
      ::encode(subusers, bl);
      ::encode(suspended, bl);
      ::encode(swift_keys, bl);
+     ::encode(max_buckets, bl);
+     ::encode(caps, bl);
      ENCODE_FINISH(bl);
   }
   void decode(bufferlist::iterator& bl) {
-     DECODE_START_LEGACY_COMPAT_LEN_32(9, 9, 9, bl);
+     DECODE_START_LEGACY_COMPAT_LEN_32(11, 9, 9, bl);
      if (struct_v >= 2) ::decode(auid, bl);
      else auid = CEPH_AUTH_UID_DEFAULT;
      string access_key;
@@ -397,6 +431,14 @@ struct RGWUserInfo
     if (struct_v >= 8) {
       ::decode(swift_keys, bl);
     }
+    if (struct_v >= 10) {
+      ::decode(max_buckets, bl);
+    } else {
+      max_buckets = RGW_DEFAULT_MAX_BUCKETS;
+    }
+    if (struct_v >= 11) {
+      ::decode(caps, bl);
+    }
     DECODE_FINISH(bl);
   }
   void dump(Formatter *f) const;
@@ -473,8 +515,6 @@ inline ostream& operator<<(ostream& out, const rgw_bucket b) {
   return out;
 }
 
-extern rgw_bucket rgw_root_bucket;
-
 enum RGWBucketFlags {
   BUCKET_SUSPENDED = 0x1,
 };
@@ -519,31 +559,32 @@ struct RGWBucketStats
 struct req_state;
 
 struct RGWEnv;
-struct FCGX_Request;
+
+class RGWClientIO;
 
 /** Store all the state necessary to complete and respond to an HTTP request*/
 struct req_state {
    CephContext *cct;
-   FCGX_Request *fcgx;
+   RGWClientIO *cio;
    http_op op;
    bool content_started;
    int format;
    ceph::Formatter *formatter;
    string decoded_uri;
    string request_uri;
+   string script_uri;
    string request_params;
    const char *host;
    const char *method;
    const char *length;
    uint64_t content_length;
-   const char *content_type;
+   map<string, string> generic_attrs;
    struct rgw_err err;
    bool expect_cont;
    bool header_ended;
-   uint64_t bytes_sent; // bytes sent as a response, excluding header
-   uint64_t bytes_received; // data received
    uint64_t obj_size;
-   bool should_log;
+   bool enable_ops_log;
+   bool enable_usage_log;
    uint32_t perm_mask;
    utime_t header_time;
 
@@ -552,8 +593,6 @@ struct req_state {
    const char *bucket_name;
    const char *object;
 
-   const char *host_bucket;
-
    rgw_bucket bucket;
    string bucket_name_str;
    string object_str;
@@ -573,8 +612,8 @@ struct req_state {
    int prot_flags;
 
    const char *os_auth_token;
-   char *os_user;
-   char *os_groups;
+   string swift_user;
+   string swift_groups;
 
    utime_t time;
 
@@ -598,6 +637,8 @@ struct RGWObjEnt {
   string etag;
   string content_type;
 
+  RGWObjEnt() : size(0), mtime(0) {}
+
   void clear() { // not clearing etag
     name="";
     size = 0;
@@ -661,35 +702,6 @@ struct RGWBucketEnt {
 };
 WRITE_CLASS_ENCODER(RGWBucketEnt)
 
-struct RGWUploadPartInfo {
-  uint32_t num;
-  uint64_t size;
-  string etag;
-  utime_t modified;
-
-  RGWUploadPartInfo() : num(0), size(0) {}
-
-  void encode(bufferlist& bl) const {
-    ENCODE_START(2, 2, bl);
-    ::encode(num, bl);
-    ::encode(size, bl);
-    ::encode(etag, bl);
-    ::encode(modified, bl);
-    ENCODE_FINISH(bl);
-  }
-  void decode(bufferlist::iterator& bl) {
-    DECODE_START_LEGACY_COMPAT_LEN(2, 2, 2, bl);
-    ::decode(num, bl);
-    ::decode(size, bl);
-    ::decode(etag, bl);
-    ::decode(modified, bl);
-    DECODE_FINISH(bl);
-  }
-  void dump(Formatter *f) const;
-  static void generate_test_instances(list<RGWUploadPartInfo*>& o);
-};
-WRITE_CLASS_ENCODER(RGWUploadPartInfo)
-
 class rgw_obj {
   std::string orig_obj;
   std::string orig_key;
@@ -983,13 +995,17 @@ static inline const char *rgw_obj_category_name(RGWObjCategory category)
   return "unknown";
 }
 
+extern string rgw_string_unquote(const string& s);
 /** time parsing */
 extern int parse_time(const char *time_str, time_t *time);
 extern bool parse_rfc2616(const char *s, struct tm *t);
+extern bool parse_iso8601(const char *s, struct tm *t);
+extern int parse_date(const string& date, uint64_t *epoch, string *out_date = NULL, string *out_time = NULL);
 
 /** Check if the req_state's user has the necessary permissions
  * to do the requested action */
 extern bool verify_bucket_permission(struct req_state *s, int perm);
+extern bool verify_object_permission(struct req_state *s, RGWAccessControlPolicy *bucket_acl, RGWAccessControlPolicy *object_acl, int perm);
 extern bool verify_object_permission(struct req_state *s, int perm);
 /** Convert an input URL into a sane object name
  * by converting %-escaped strings into characters, etc*/
