@@ -17,6 +17,8 @@
 # include "fcgiapp.h"
 #endif
 
+#include "rgw_fcgi.h"
+
 #include "common/ceph_argparse.h"
 #include "global/global_init.h"
 #include "global/signal_handler.h"
@@ -25,15 +27,22 @@
 #include "common/WorkQueue.h"
 #include "common/Timer.h"
 #include "common/Throttle.h"
+#include "include/str_list.h"
 #include "rgw_common.h"
 #include "rgw_rados.h"
 #include "rgw_acl.h"
 #include "rgw_user.h"
 #include "rgw_op.h"
 #include "rgw_rest.h"
+#include "rgw_rest_s3.h"
+#include "rgw_rest_swift.h"
+#include "rgw_rest_admin.h"
+#include "rgw_rest_usage.h"
+#include "rgw_swift_auth.h"
 #include "rgw_swift.h"
 #include "rgw_log.h"
 #include "rgw_tools.h"
+#include "rgw_resolve.h"
 
 #include <map>
 #include <string>
@@ -52,21 +61,13 @@ static sighandler_t sighandler_usr1;
 static sighandler_t sighandler_alrm;
 static sighandler_t sighandler_term;
 
+class RGWProcess;
+
+static RGWProcess *pprocess = NULL;
+
 
 #define SOCKET_BACKLOG 1024
 
-static void godown_handler(int signum)
-{
-  FCGX_ShutdownPending();
-  signal(signum, sighandler_usr1);
-  alarm(5);
-}
-
-static void godown_alarm(int signum)
-{
-  _exit(0);
-}
-
 struct RGWRequest
 {
   FCGX_Request fcgx;
@@ -109,10 +110,6 @@ struct RGWRequest
     if (s->method && req_str.size() == 0) {
       req_str = s->method;
       req_str.append(" ");
-      if (s->host_bucket) {
-        req_str.append(s->host_bucket);
-        req_str.append("/");
-      }
       req_str.append(s->request_uri);
     }
     utime_t t = ceph_clock_now(g_ceph_context) - ts;
@@ -121,9 +118,13 @@ struct RGWRequest
 };
 
 class RGWProcess {
+  RGWRados *store;
+  OpsLogSocket *olog;
   deque<RGWRequest *> m_req_queue;
   ThreadPool m_tp;
   Throttle req_throttle;
+  RGWREST *rest;
+  int sock_fd;
 
   struct RGWWQ : public ThreadPool::WorkQueue<RGWRequest> {
     RGWProcess *process;
@@ -178,25 +179,45 @@ class RGWProcess {
   uint64_t max_req_id;
 
 public:
-  RGWProcess(CephContext *cct, int num_threads)
-    : m_tp(cct, "RGWProcess::m_tp", num_threads),
-      req_throttle(num_threads * 2),
+  RGWProcess(CephContext *cct, RGWRados *rgwstore, OpsLogSocket *_olog, int num_threads, RGWREST *_rest)
+    : store(rgwstore), olog(_olog), m_tp(cct, "RGWProcess::m_tp", num_threads),
+      req_throttle(cct, "rgw_ops", num_threads * 2),
+      rest(_rest), sock_fd(-1),
       req_wq(this, g_conf->rgw_op_thread_timeout,
 	     g_conf->rgw_op_thread_suicide_timeout, &m_tp),
       max_req_id(0) {}
   void run();
   void handle_request(RGWRequest *req);
+
+  void close_fd() {
+    if (sock_fd >= 0)
+      close(sock_fd);
+  }
 };
 
 void RGWProcess::run()
 {
-  int s = 0;
+  sock_fd = 0;
   if (!g_conf->rgw_socket_path.empty()) {
     string path_str = g_conf->rgw_socket_path;
+
+    /* this is necessary, as FCGX_OpenSocket might not return an error, but rather ungracefully exit */
+    int fd = open(path_str.c_str(), O_CREAT, 0644);
+    if (fd < 0) {
+      int err = errno;
+      /* ENXIO is actually expected, we'll get that if we try to open a unix domain socket */
+      if (err != ENXIO) {
+        dout(0) << "ERROR: cannot create socket: path=" << path_str << " error=" << cpp_strerror(err) << dendl;
+        return;
+      }
+    } else {
+      close(fd);
+    }
+
     const char *path = path_str.c_str();
-    s = FCGX_OpenSocket(path, SOCKET_BACKLOG);
-    if (s < 0) {
-      dout(0) << "ERROR: FCGX_OpenSocket (" << path << ") returned " << s << dendl;
+    sock_fd = FCGX_OpenSocket(path, SOCKET_BACKLOG);
+    if (sock_fd < 0) {
+      dout(0) << "ERROR: FCGX_OpenSocket (" << path << ") returned " << sock_fd << dendl;
       return;
     }
     if (chmod(path, 0777) < 0) {
@@ -210,11 +231,15 @@ void RGWProcess::run()
     RGWRequest *req = new RGWRequest;
     req->id = ++max_req_id;
     dout(10) << "allocated request req=" << hex << req << dec << dendl;
-    FCGX_InitRequest(&req->fcgx, s, 0);
+    FCGX_InitRequest(&req->fcgx, sock_fd, 0);
     req_throttle.get(1);
     int ret = FCGX_Accept_r(&req->fcgx);
-    if (ret < 0)
+    if (ret < 0) {
+      delete req;
+      dout(0) << "ERROR: FCGX_Accept_r returned " << ret << dendl;
+      req_throttle.put(1);
       break;
+    }
 
     req_wq.queue(req);
   }
@@ -222,18 +247,33 @@ void RGWProcess::run()
   m_tp.stop();
 }
 
-static int call_log_intent(void *ctx, rgw_obj& obj, RGWIntentEvent intent)
+static void godown_handler(int signum)
+{
+  FCGX_ShutdownPending();
+  pprocess->close_fd();
+  signal(signum, sighandler_usr1);
+  uint64_t secs = g_ceph_context->_conf->rgw_exit_timeout_secs;
+  if (secs)
+    alarm(secs);
+}
+
+static void godown_alarm(int signum)
+{
+  _exit(0);
+}
+
+static int call_log_intent(RGWRados *store, void *ctx, rgw_obj& obj, RGWIntentEvent intent)
 {
   struct req_state *s = (struct req_state *)ctx;
-  return rgw_log_intent(s, obj, intent);
+  return rgw_log_intent(store, s, obj, intent);
 }
 
 void RGWProcess::handle_request(RGWRequest *req)
 {
   FCGX_Request *fcgx = &req->fcgx;
-  RGWRESTMgr rest;
   int ret;
   RGWEnv rgw_env;
+  RGWFCGX client_io(fcgx);
 
   req->log_init();
 
@@ -243,21 +283,21 @@ void RGWProcess::handle_request(RGWRequest *req)
   rgw_env.init(g_ceph_context, fcgx->envp);
 
   struct req_state *s = req->init_state(g_ceph_context, &rgw_env);
-  s->obj_ctx = rgwstore->create_context(s);
-  rgwstore->set_intent_cb(s->obj_ctx, call_log_intent);
+  s->obj_ctx = store->create_context(s);
+  store->set_intent_cb(s->obj_ctx, call_log_intent);
 
   req->log(s, "initializing");
 
   RGWOp *op = NULL;
   int init_error = 0;
-  RGWHandler *handler = rest.get_handler(s, fcgx, &init_error);
+  RGWHandler *handler = rest->get_handler(store, s, &client_io, &init_error);
   if (init_error != 0) {
     abort_early(s, init_error);
     goto done;
   }
 
   req->log(s, "getting op");
-  op = handler->get_op();
+  op = handler->get_op(store);
   if (!op) {
     abort_early(s, -ERR_METHOD_NOT_ALLOWED);
     goto done;
@@ -303,19 +343,22 @@ void RGWProcess::handle_request(RGWRequest *req)
 
   req->log(s, "executing");
   op->execute();
+  op->complete();
 done:
-  rgw_log_op(s);
+  rgw_log_op(store, s, (op ? op->name() : "unknown"), olog);
 
   int http_ret = s->err.http_ret;
 
   req->log_format(s, "http status=%d", http_ret);
 
-  handler->put_op(op);
-  rgwstore->destroy_context(s->obj_ctx);
+  if (handler)
+    handler->put_op(op);
+  rest->put_handler(handler);
+  store->destroy_context(s->obj_ctx);
   FCGX_Finish_r(fcgx);
-  delete req;
 
   dout(1) << "====== req done req=" << hex << req << dec << " http_status=" << http_ret << " ======" << dendl;
+  delete req;
 }
 
 class C_InitTimeout : public Context {
@@ -345,6 +388,7 @@ int main(int argc, const char **argv)
   /* alternative default for module */
   vector<const char *> def_args;
   def_args.push_back("--debug-rgw=20");
+  def_args.push_back("--keyring=$rgw_data/keyring");
 
   vector<const char*> args;
   argv_to_vec(argc, argv, args);
@@ -352,32 +396,13 @@ int main(int argc, const char **argv)
   global_init(&def_args, args, CEPH_ENTITY_TYPE_CLIENT, CODE_ENVIRONMENT_DAEMON,
 	      CINIT_FLAG_UNPRIVILEGED_DAEMON_DEFAULTS);
 
-  pid_t childpid = 0;
   if (g_conf->daemonize) {
     if (g_conf->rgw_socket_path.empty()) {
       cerr << "radosgw: must specify 'rgw socket path' to run as a daemon" << std::endl;
       exit(1);
     }
 
-    g_ceph_context->_log->stop();
-
-    childpid = fork();
-    if (childpid) {
-      // i am the parent
-      cout << "radosgw daemon started with pid " << childpid << std::endl;
-      exit(0);
-    }
-
-    // i am child
-    close(0);
-    close(1);
-    close(2);
-    int r = chdir(g_conf->chdir.c_str());
-    if (r < 0) {
-      dout(0) << "weird, i couldn't chdir to '" << g_conf->chdir << "'" << dendl;
-    }
-
-    g_ceph_context->_log->start();
+    global_init_daemonize(g_ceph_context, 0);
   }
   Mutex mutex("main");
   SafeTimer init_timer(g_ceph_context, mutex);
@@ -389,6 +414,9 @@ int main(int argc, const char **argv)
   common_init_finish(g_ceph_context);
 
   rgw_tools_init(g_ceph_context);
+
+  rgw_init_resolver();
+  rgw_rest_init(g_ceph_context);
   
   curl_global_init(CURL_GLOBAL_ALL);
   
@@ -402,28 +430,91 @@ int main(int argc, const char **argv)
 
   sighandler_term = signal(SIGTERM, godown_alarm);
   
-  RGWStoreManager store_manager;
-  
-  if (!store_manager.init(g_ceph_context)) {
+
+  int r = 0;
+  RGWRados *store = RGWStoreManager::get_storage(g_ceph_context, true);
+  if (!store) {
     derr << "Couldn't init storage provider (RADOS)" << dendl;
-    return EIO;
+    r = EIO;
   }
-  
-  int r = rgw_perf_start(g_ceph_context);
-  if (r < 0)
-    return 1;
+  if (!r)
+    r = rgw_perf_start(g_ceph_context);
 
   mutex.Lock();
   init_timer.cancel_all_events();
+  init_timer.shutdown();
   mutex.Unlock();
 
-  RGWProcess process(g_ceph_context, g_conf->rgw_thread_pool_size);
-  process.run();
+  if (r) 
+    return 1;
+
+  rgw_log_usage_init(g_ceph_context, store);
+
+  RGWREST rest;
+
+  list<string> apis;
+  bool do_swift = false;
+
+  get_str_list(g_conf->rgw_enable_apis, apis);
+
+  map<string, bool> apis_map;
+  for (list<string>::iterator li = apis.begin(); li != apis.end(); ++li) {
+    apis_map[*li] = true;
+  }
+
+  if (apis_map.count("s3") > 0)
+    rest.register_default_mgr(new RGWRESTMgr_S3);
+
+  if (apis_map.count("swift") > 0) {
+    do_swift = true;
+    swift_init(g_ceph_context);
+    rest.register_resource(g_conf->rgw_swift_url_prefix, new RGWRESTMgr_SWIFT);
+  }
+
+  if (apis_map.count("swift_auth") > 0)
+    rest.register_resource(g_conf->rgw_swift_auth_entry, new RGWRESTMgr_SWIFT_Auth);
+
+  if (apis_map.count("admin") > 0) {
+    RGWRESTMgr_Admin *admin_resource = new RGWRESTMgr_Admin;
+    admin_resource->register_resource("usage", new RGWRESTMgr_Usage);
+    rest.register_resource(g_conf->rgw_admin_entry, admin_resource);
+  }
+
+  OpsLogSocket *olog = NULL;
+
+  if (!g_conf->rgw_ops_log_socket_path.empty()) {
+    olog = new OpsLogSocket(g_ceph_context, g_conf->rgw_ops_log_data_backlog);
+    olog->init(g_conf->rgw_ops_log_socket_path);
+  }
+
+  pprocess = new RGWProcess(g_ceph_context, store, olog, g_conf->rgw_thread_pool_size, &rest);
+
+  pprocess->run();
+
+  delete pprocess;
+
+  if (do_swift) {
+    swift_finalize();
+  }
+
+  rgw_log_usage_finalize();
+
+  delete olog;
 
   rgw_perf_stop(g_ceph_context);
 
   unregister_async_signal_handler(SIGHUP, sighup_handler);
 
+  RGWStoreManager::close_storage(store);
+
+  rgw_tools_cleanup();
+  curl_global_cleanup();
+  g_ceph_context->put();
+
+  shutdown_async_signal_handler();
+
+  ceph::crypto::shutdown();
+
   return 0;
 }
 
